/*---------------------------------------------------------------------------*\
  =========                 |
  \\      /  F ield         | OpenFOAM: The Open Source CFD Toolbox
   \\    /   O peration     |
    \\  /    A nd           | www.openfoam.com
     \\/     M anipulation  |
-------------------------------------------------------------------------------
    Copyright (C) 2011-2016 OpenFOAM Foundation
    Copyright (C) 2019 OpenCFD Ltd.
-------------------------------------------------------------------------------
License
    This file is part of OpenFOAM.

    OpenFOAM is free software: you can redistribute it and/or modify it
    under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    OpenFOAM is distributed in the hope that it will be useful, but WITHOUT
    ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
    FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
    for more details.

    You should have received a copy of the GNU General Public License
    along with OpenFOAM.  If not, see <http://www.gnu.org/licenses/>.

Application
    icoLduAddressingFoam

Description
	A special version of icoFoam solver created for educational purposes to
	understand and explain the LDU addressing mechanism in OpenFOAM. The
	UEqnWithPressure equation object is created which is the full momentum
	equation. The non-zero values of the coefficient matrix of this equation,
	i.e., lower(), upper(), diag(), and source() coefficients, as well as the
	LDU addressing (lowerAddr(), upperAddr(), and ownerStartAddr()) are
	extracted and written to the output.

	Then, these coefficients and addressing are used to construct the full
	coefficient matrix of the x-component of the momentum equation using the
	simpleMatrix class (N by N, N is the number of cells). Subsequently, the
	effects of boundary conditions are added to the diagonal and source terms.
	The full linear system is then solved and it is shown that it produces
	identical results to the icoFoam results for Ux.

Author
	Saeed Salehi, saeed.salehi@chalmers.se
\*---------------------------------------------------------------------------*/

#include "fvCFD.H"
#include "pisoControl.H"
#include "simpleMatrix.H"
#include "scalarField.H"
// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * //

int main(int argc, char *argv[])
{
    argList::addNote
    (
        "Transient solver for incompressible, laminar flow"
        " of Newtonian fluids."
    );

    #include "postProcess.H"

    #include "addCheckCaseOptions.H"
    #include "setRootCaseLists.H"
    #include "createTime.H"
    #include "createMesh.H"

    pisoControl piso(mesh);

    #include "createFields.H"
    #include "initContinuityErrs.H"

    // * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * //

    Info<< "\nStarting time loop\n" << endl;

    while (runTime.loop())
    {
        Info<< "Time = " << runTime.timeName() << nl << endl;

        #include "CourantNo.H"

	// Momentum predictor
        fvVectorMatrix UEqn
        (
            //fvm::ddt(U)
          fvm::div(phi, U)
          //- fvm::laplacian(nu, U)
        );

        fvVectorMatrix UEqnWithPressure(UEqn == -fvc::grad(p));

        if (piso.momentumPredictor())
        {
            solve(UEqnWithPressure);
        }

        labelUList lAdd = UEqnWithPressure.lduAddr().lowerAddr();
        labelUList uAdd = UEqnWithPressure.lduAddr().upperAddr();
    	labelUList oAdd = UEqnWithPressure.lduAddr().ownerStartAddr();
        scalarField lower = UEqnWithPressure.lower();
        scalarField upper = UEqnWithPressure.upper();
        scalarField diag = UEqnWithPressure.diag();
        vectorField source = UEqnWithPressure.source();

        Info << "lowerAddr:" << lAdd << nl << endl;
        Info << "upperAddr:" << uAdd << nl << endl;
        Info << "ownerStartAddr:" << nl << oAdd << nl << endl;
        Info << "lower:" << lower << nl << endl;
        Info << "upper:" << upper << nl << endl;
        Info << "diag:" << nl << diag << nl << endl;
        Info << "source:" << nl << source << nl << endl;
        //Info << UEqn  << endl;

        #include "createCoeffMatrix.H"


        // --- PISO loop
        while (piso.correct())
        {
            volScalarField rAU(1.0/UEqn.A());
            volVectorField HbyA(constrainHbyA(rAU*UEqn.H(), U, p));
            surfaceScalarField phiHbyA
            (
                "phiHbyA",
                fvc::flux(HbyA)
              + fvc::interpolate(rAU)*fvc::ddtCorr(U, phi)
            );

            adjustPhi(phiHbyA, U, p);

            // Update the pressure BCs to ensure flux consistency
            constrainPressure(p, U, phiHbyA, rAU);

            // Non-orthogonal pressure corrector loop
            while (piso.correctNonOrthogonal())
            {
                // Pressure corrector

                fvScalarMatrix pEqn
                (
                    fvm::laplacian(rAU, p) == fvc::div(phiHbyA)
                );

                pEqn.setReference(pRefCell, pRefValue);

                pEqn.solve(mesh.solver(p.select(piso.finalInnerIter())));

                if (piso.finalNonOrthogonalIter())
                {
                    phi = phiHbyA - pEqn.flux();
                }
            }

            #include "continuityErrs.H"

            U = HbyA - rAU*fvc::grad(p);
            U.correctBoundaryConditions();
        }

        runTime.write();

        runTime.printExecutionTime(Info);
    }

    Info<< "End\n" << endl;

    return 0;
}


// ************************************************************************* //
